import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as st
from sympy.parsing.sympy_parser import parse_expr


def iniciacion_variables(info):
    """
    Defino las matrices dónde iré almacenando los valores de las variables y
    de sus índices de sensibilidad (alpha) obtenidos en cada iteración.

    Igualmente, defino el vector dónde iré almacenando el valor de la probabilidad
    de fallo (pf), el índice de fiabilidad (beta), y el valor de la ecuación de
    verificación (G) obtenidos en cada iteración.

    Es necesario definir el punto crítico inicial como una lista con un número de
    elementos igual al número de variables que hay en nuestro problema.
    """
    resultados = dict()

    for var_ in info["variables"]:
        resultados[var_] = []
        resultados["alpha_" + var_] = []

    output_names = ["yi", "dh_dyi", "medias", "desv_tip"]

    for name_ in output_names:
        resultados[name_] = {}

    output_names = ["pf", "beta", "G"]
    for name_ in output_names:
        resultados[name_] = []

    # Ecuación de verificación
    ec_eval = parse_expr(info["ec_verificacion"])

    return resultados, ec_eval


def trans_gaussiana(pto_aprox, info, var):
    """
    Aproxima cualquier función de distribución a una distribución
    Gaussiana en un entorno cercano a un punto.
    """
    for ind_, var_ in enumerate(info["variables"]):
        dist = info["modelos_probabilidad"][ind_]
        params = info["params_mod_prob"][ind_]
        if dist != "norm":
            Fx0 = getattr(st, dist).cdf(pto_aprox, *params)
            fx0 = getattr(st, dist).pdf(pto_aprox, *params)

            Phim1_Fx0 = st.norm.ppf(Fx0)
            phi = st.norm.pdf(Phim1_Fx0)
            var["desv_tip"][var_] = phi / fx0
            var["medias"][var_] = pto_aprox - Phim1_Fx0 * var["desv_tip"][var_]
        else:
            var["desv_tip"][var_] = params[1]
            var["medias"][var_] = params[0]

    return var


def normalizacion(noiter, var, info):
    """
    Transforma las distribuciones gaussianas en distribuciones gausianas de
    parámetros [0,1]
    """
    for ind_, var_ in enumerate(info["variables"]):
        var["yi"][var_] = (var[var_][noiter] - var["medias"][var_]) / var["desv_tip"][
            var_
        ]
    return var


def derivada(ec_eval, info, var):
    import sympy as sp

    """
    Cálculo de las derivadas de la función ec respecto a la variable var
    """
    for ind_, var_ in enumerate(info["variables"]):
        var["dh_dyi"][var_] = sp.diff(ec_eval, var_)
    return var


def evalua(ec, par, var):
    """
    Evaluación de la ecuación sustituyendo los valores de los parámetros par de
    un diccionario por su valor.
    """
    ec_eval = ec.subs(par)
    ec_eval = ec_eval.subs(var)

    return ec_eval


def coeficientes(G, n_variables, var, noiter):
    """
    Cálculo de los coeficientes de la función de verificación linealizada
    """
    a0 = G

    for i in n_variables:
        a0 = a0 - var["dh_dyi"][i] * var["yi"][i]
    ai = {}
    for i in n_variables:
        ai[i] = var["dh_dyi"][i]

    aii = np.asarray([float(a) for a in ai.values()])
    raiz = np.sqrt(np.sum(aii ** 2))

    beta = abs(a0) / raiz

    # Cálculo de los índices de sensibilidad de g
    sensibilidad = []
    for i in n_variables:
        sensibilidad.append(ai[i] / raiz)

    # Cálculo de las coordenadas del punto crítico de g
    y_im = {}
    y_is = {}
    for i in n_variables:
        y_im[i] = -a0 * ai[i] / raiz ** 2
        # Calculo de las coordenadas de la primera aproximación al punto crítico de g0
        y_is[i] = float(var["medias"][i] + var["desv_tip"][i] * y_im[i])

    coef = {"beta": beta, "alphas": sensibilidad, "y_is": y_is}

    # for i in range(n):
    for ind_, var_ in enumerate(n_variables):
        var[var_].append(coef["y_is"][var_])
        var["alpha_" + n_variables[ind_]].append(coef["alphas"][ind_])
    var["beta"].append(float(coef["beta"]))
    var["pf"].append(st.norm.cdf(-var["beta"][noiter]))

    return var, coef


def panels_graph(var, info):

    # Parte gráfica

    # Dibujamos los resultados
    fig, axs = plt.subplots(2, 4, figsize=(17, 4))
    ax = axs.flatten()

    for ind_, var_ in enumerate(info["variables"]):
        ax[ind_].plot(var[var_], ".-")
        ax[ind_].set_ylabel(var_)
        ax[ind_].set_xlabel("Iteration")

    for var_ in info["variables"]:
        ax[4].set_ylabel("Alpha")
        ax[4].plot(var["alpha_" + var_], ".-", label="Alpha " + var_)
        ax[4].set_xlabel("Iteration")
        ax[4].legend(loc="best")

    ax[5].plot(var["pf"], ".-")
    ax[5].set_ylabel("pf")
    ax[5].set_xlabel("Iteration")

    ax[6].plot(var["beta"], ".-")
    ax[6].set_ylabel("beta")
    ax[6].set_xlabel("Iteration")

    ax[7].plot(var["G"], ".-")
    ax[7].set_ylabel("G")
    ax[7].set_xlabel("Iteration")
    plt.show()
    return
